"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import fnmatch

from typing import List
from typing import Dict
from typing import Optional
from typing import Union

from ..tfile import Torrent
from ..tfile import is_torrent_hash

from ..cli import Log


# =====
def load_from_dir(path: str, name_filter: str, precalculate_hashes: bool, log: Log) -> Dict[str, Optional[Torrent]]:
    if not log.isatty():
        log.info("Loading torrents from {cyan}%s/{yellow}%s{reset} ...", (path, name_filter))

    torrents: Dict[str, Optional[Torrent]] = {}
    for name in log.progress(
        sorted(name for name in os.listdir(path) if fnmatch.fnmatch(name, name_filter)),
        ("Loading torrents from {cyan}%s/{yellow}%s{reset}", (path, name_filter)),
        ("Loaded {magenta}%d{reset} torrents from {cyan}%s/{yellow}%s{reset}", (lambda: len(torrents), path, name_filter)),
    ):
        file_path = os.path.abspath(os.path.join(path, name))
        try:
            torrents[name] = Torrent(path=file_path)
            if precalculate_hashes:
                torrents[name].get_hash()  # type: ignore
        except ValueError:
            log.error("Found broken torrent: {cyan}%s/{yellow}%s{reset}", (path, name))
            torrents[name] = None
        except Exception:
            log.error("Can't process torrent: {cyan}%s/{yellow}%s{reset}", (path, name))
            raise

    if not log.isatty():
        log.info("Loaded {magenta}%d{reset} torrents from {cyan}%s/{yellow}%s{reset}",
                 (len(torrents), path, name_filter))
    return torrents


def by_hash(torrents: Dict[str, Optional[Torrent]]) -> Dict[str, Torrent]:
    return {
        torrent.get_hash(): torrent
        for torrent in filter(None, torrents.values())
    }


def by_hash_with_dups(torrents: Dict[str, Optional[Torrent]]) -> Dict[str, List[Torrent]]:
    with_dups: Dict[str, List[Torrent]] = {}  # noqa: E701
    for torrent in filter(None, torrents.values()):
        with_dups.setdefault(torrent.get_hash(), [])
        with_dups[torrent.get_hash()].append(torrent)
    return with_dups


# =====
def find_torrents(path: str, items: str) -> List[Torrent]:
    return [_find_torrent_or_hash(path, item, False) for item in items]  # type: ignore


def find_torrents_or_hashes(path: str, items: str) -> List[Union[Torrent, str]]:
    return [_find_torrent_or_hash(path, item, True) for item in items]


def _find_torrent_or_hash(path: str, item: str, pass_hash: bool) -> Union[Torrent, str]:
    if os.path.exists(item):
        return Torrent(path=os.path.abspath(item))
    if os.path.sep not in item:
        full_path = os.path.join(path, item)
        if os.path.exists(full_path):
            return Torrent(path=full_path)
    if pass_hash and is_torrent_hash(item.strip()):
        return item.strip()
    raise RuntimeError("Can't find torrent: {}".format(item))



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    atom.py -- produce atom feed file of recent torrent updates
    Copyright (C) 2017  Pavel Pletenev <cpp.create@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import getpass
import pwd
import grp
import traceback
import time

from typing import List
from typing import Dict
from typing import Any

import yaml

from ...optconf import Option
from ...optconf.converters import as_path
from ...optconf.converters import as_path_or_empty

from . import ResultsType
from . import BaseConfetti
from . import templated


# =====
def get_uid(user: str) -> int:
    return pwd.getpwnam(user)[2]


def get_gid(group: str) -> int:
    return grp.getgrnam(group)[2]


def get_user_groups(user: str) -> List[int]:
    groups = [
        group.gr_name
        for group in grp.getgrall()
        if user in group.gr_mem
    ]
    gid = pwd.getpwnam(user).pw_gid
    groups.append(grp.getgrgid(gid).gr_name)
    return [grp.getgrnam(group).gr_gid for group in groups]


class UserError(Exception):
    pass


# =====
class Plugin(BaseConfetti):  # pylint: disable=too-many-instance-attributes
    PLUGIN_NAME = "atom"

    def __init__(  # pylint: disable=super-init-not-called,too-many-arguments
        self,
        history_path: str,
        path: str,
        url: str,
        user: str,
        group: str,
        template: str,
        html: bool,
        **kwargs: Any,
    ) -> None:

        self._init_bases(**kwargs)

        self._history_path = history_path
        self._path = path
        self._url = url
        self._uid = (get_uid(user) if user else -1)
        self._gid = (get_gid(group) if group else -1)
        if self._gid > -1:
            if self._gid not in get_user_groups(getpass.getuser()):
                raise UserError(
                    "I wouldn't be able to edit {path} with current user if I chown it for "
                    "uid={uid} and gid={gid}".format(path=path, uid=self._uid, gid=self._gid),
                )
        self._template_path = template
        self._html = html

    @classmethod
    def get_options(cls) -> Dict[str, Option]:
        return cls._get_merged_options({
            "history_path": Option(default="emonoda_history.yaml", type=as_path, help="History path"),
            "path":         Option(default="atom.xml", type=as_path, help="Atom path"),
            "url":          Option(default="http://localhost/", help="Feed server url"),
            "user":         Option(default="", help="Server user"),
            "group":        Option(default="", help="Server user group"),
            "template":     Option(default="", type=as_path_or_empty, help="Mako template file name"),
            "html":         Option(default=True, help="HTML or plaintext feed")
        })

    # ===

    def send_results(self, source: str, results: ResultsType) -> None:
        results_set: List[ResultsType] = []
        try:
            with open(self._history_path) as history_file:
                results_set = yaml.load(history_file)
        except Exception:
            traceback.print_exc()
        results["ctime"] = time.time()  # type: ignore
        results_set.insert(0, results)
        results_set = results_set[:20]
        with open(self._path, "w") as atom_file:
            atom_file.write(templated(
                name=(self._template_path if self._template_path else "atom.{ctype}.{source}.mako").format(
                    ctype=("html" if self._html else "plain"),
                    source=source,
                    results_set=results_set,
                ),
                built_in=(not self._template_path),
                source=source,
                results_set=results_set,
                settings=dict(url=self._url)
            ))
        os.chmod(self._path, 0o664)
        os.chown(self._path, self._uid, self._gid)
        with open(self._history_path, "w") as history_file:
            history_file.write(yaml.dump(results_set))
        del results["ctime"]



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import importlib
import functools
import os

from typing import Tuple
from typing import List
from typing import Dict
from typing import Optional
from typing import Type
from typing import Any

from ..optconf import Option


# =====
class BasePlugin:
    PLUGIN_NAME = ""

    def __init__(self, **_: Any) -> None:
        pass

    @classmethod
    def get_options(cls) -> Dict[str, Option]:
        return {}

    @classmethod
    def get_bases(cls) -> "List[Type[BasePlugin]]":
        return cls.__get_bases(cls.__mro__)

    def _init_bases(self, **kwargs: Any) -> None:
        assert self.PLUGIN_NAME
        for parent in self.__get_bases(self.__class__.__mro__):
            parent.__init__(self, **kwargs)  # type: ignore

    @classmethod
    def _get_merged_options(cls, params: Optional[Dict[str, Option]]=None) -> Dict[str, Option]:
        merged: Dict[str, Option] = {}
        for parent in cls.__get_bases(cls.__mro__):
            merged.update(parent.get_options())
        merged.update(params or {})
        return merged

    @staticmethod
    def __get_bases(mro: Tuple[Type, ...]) -> "List[Type[BasePlugin]]":
        return [
            cls for cls in mro
            if issubclass(cls, BasePlugin)
        ][1:]


# =====
@functools.lru_cache()
def get_classes(sub: str) -> Dict[str, Type[BasePlugin]]:
    classes: Dict[str, Type[BasePlugin]] = {}  # noqa: E701
    sub_path = os.path.join(os.path.dirname(__file__), sub)
    for file_name in os.listdir(sub_path):
        if not file_name.startswith("__") and file_name.endswith(".py"):
            module_name = file_name[:-3]
            module = importlib.import_module("emonoda.plugins.{}.{}".format(sub, module_name))
            plugin_class = getattr(module, "Plugin")
            classes[plugin_class.PLUGIN_NAME] = plugin_class
    return classes



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import re

from datetime import datetime

from typing import Dict
from typing import Any

from ...optconf import Option

from ...tfile import Torrent

from . import WithLogin
from . import WithCheckTime
from . import WithFetchByDownloadId
from . import WithStat


# =====
class Plugin(WithLogin, WithCheckTime, WithFetchByDownloadId, WithStat):
    PLUGIN_NAME = "booktracker.org"

    _SITE_VERSION = 2

    _SITE_FINGERPRINT_URL = "https://booktracker.org"
    _SITE_FINGERPRINT_TEXT = "<meta name=\"google-site-verification\" content=\"L85kL3qg3y9JS1ER3BNhcpcqdDZBgrxzZpBm6Jzb1iQ\" />"

    _COMMENT_REGEXP = re.compile(r"https?://booktracker\.org(:443)?/viewtopic\.php\?p=(?P<torrent_id>\d+)")

    _TIMEZONE_URL = "https://booktracker.org/profile.php?mode=editprofile"
    _TIMEZONE_REGEXP = re.compile(r"<option value=\"[\.\d+-]\" selected=\"selected\">(?P<timezone>GMT [+-] [\d\.]+)[\s<\(]")
    _TIMEZONE_PREFIX = "Etc/"

    _DOWNLOAD_ID_URL = "https://booktracker.org/viewtopic.php?p={torrent_id}"
    _DOWNLOAD_ID_REGEXP = re.compile(r"<a href=\"download\.php\?id=(?P<download_id>\d+)\" class=\"\">")
    _DOWNLOAD_URL = "https://booktracker.org/download.php?id={download_id}"

    _STAT_URL = "https://booktracker.org()/viewtopic.php?p={torrent_id}"
    _STAT_SEEDERS_REGEXP = re.compile(r"<b>Раздают:\s+(?P<seeders>\d+)</b></span> &#0183;")
    _STAT_LEECHERS_REGEXP = re.compile(r"<span class=\"leechmed\" ><b>Качают:\s+(?P<leechers>\d+)</b></span>")

    # ===

    def __init__(self, **kwargs: Any) -> None:  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)

    @classmethod
    def get_options(cls) -> Dict[str, Option]:
        return cls._get_merged_options()

    def fetch_time(self, torrent: Torrent) -> int:
        self._assert_match(torrent)
        date = self._assert_logic_re_search(
            regexp=re.compile(r"Зарегистрирован &nbsp;\s*\[ <span title=\"[\w\s]+\">"
                              r"(\d\d\d\d-\d\d-\d\d \d\d:\d\d)</span> ]"),
            text=self._decode(self._read_url(torrent.get_comment())),
            msg="Upload date not found",
        ).group(1)
        date += " " + datetime.now(self._tzinfo).strftime("%z")
        upload_time = int(datetime.strptime(date, "%Y-%m-%d %H:%M %z").strftime("%s"))
        return upload_time

    def login(self) -> None:
        self._login_using_post(
            url="https://booktracker.org/login.php",
            post={
                "login_username": self._encode(self._user),
                "login_password": self._encode(self._passwd),
                "login":          self._encode("Вход"),
            },
            ok_text="<b class=\"med\">{}</b></a>&nbsp; [ <a href=\"./login.php?logout=1".format(self._user),
        )



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import os
import argparse

from typing import List
from typing import Dict

from ..plugins.clients import BaseClient

from ..helpers import tcollection
from ..helpers import datacache

from ..tfile import TorrentEntryAttrs

from ..cli import Log

from .. import fmt
from .. import tools

from . import init
from . import wrap_main
from . import get_configured_log
from . import get_configured_client


# =====
def build_used_files(cache: datacache.TorrentsCache, data_roots: List[str]) -> Dict[str, TorrentEntryAttrs]:
    files: Dict[str, TorrentEntryAttrs] = dict.fromkeys(data_roots, TorrentEntryAttrs.new_dir())
    for c_attrs in cache.torrents.values():
        prefix = os.path.normpath(c_attrs.prefix)

        for (path, f_attrs) in c_attrs.files.items():
            files[os.path.join(prefix, path)] = f_attrs

        for data_root_path in data_roots:
            if prefix.startswith(data_root_path):
                parts: List[str] = list(filter(None, prefix[len(data_root_path):].split(os.path.sep)))
                for index in range(len(parts)):
                    path = os.path.join(*([data_root_path] + parts[:index + 1]))
                    files[path] = TorrentEntryAttrs.new_dir()
                break
    return files


def build_all_files(data_root_path: str, log: Log) -> Dict[str, TorrentEntryAttrs]:
    if not log.isatty():
        log.info("Scanning directory {cyan}%s{reset} ...", (data_root_path,))

    files: Dict[str, TorrentEntryAttrs] = {}
    for (prefix, _, local_files) in log.progress(
        os.walk(data_root_path),
        ("Scanning directory {cyan}%s{reset} ...", (data_root_path,)),
        ("Scanned directory {cyan}%s{reset}", (data_root_path,))
    ):
        files[tools.get_decoded_path(prefix)] = TorrentEntryAttrs.new_dir()
        for name in local_files:
            path = os.path.join(prefix, name)
            try:
                size = os.path.getsize(path)
            except FileNotFoundError:
                continue
            files[tools.get_decoded_path(path)] = TorrentEntryAttrs.new_file(size)

    if not log.isatty():
        log.info("Scanned directory {cyan}%s{reset}", (data_root_path,))
    return files


# =====
def print_orphaned_files(  # pylint: disable=too-many-locals
    cache: datacache.TorrentsCache,
    data_roots: List[str],
    ignore_orphans: List[str],
    reduce_dirs: bool,
    log_stdout: Log,
    log_stderr: Log,
) -> None:

    all_files = {}
    for data_root_path in data_roots:
        all_files.update(build_all_files(data_root_path, log_stderr))

    log_stderr.info("Transposing the cache: by-hashes -> files ...")
    used_files = build_used_files(cache, data_roots)

    files = set(all_files).difference(used_files)
    for path in list(files):
        for ignored_path in ignore_orphans:
            if path == ignored_path or path.startswith(ignored_path + os.path.sep):
                files.remove(path)
                break

    if len(files) != 0:
        log_stderr.info("Orhpaned files:")
        size = 0
        common_root = "\0"
        for path in tools.sorted_paths(files):
            size += all_files[path].size
            if reduce_dirs:
                if path.startswith(common_root + os.path.sep):
                    continue
                else:
                    common_root = (path if all_files[path].is_dir else "\0")
            line = ("{blue}D" if all_files[path].is_dir else "{magenta}F") + "{reset} %s"
            log_stdout.print(line, (path,))
        log_stderr.info("Found {red}%d{reset} orphaned files = {red}%s{reset}",
                        (len(files), fmt.format_size(size)))
    else:
        log_stderr.info("No orphaned files found")


def print_not_in_client(
    client: BaseClient,
    torrents_dir_path: str,
    name_filter: str,
    log_stdout: Log,
    log_stderr: Log,
) -> None:

    torrents = tcollection.by_hash(tcollection.load_from_dir(
        path=torrents_dir_path,
        name_filter=name_filter,
        precalculate_hashes=True,
        log=log_stderr,
    ))

    log_stderr.info("Fetching all hashes from client ...")
    client_hashes = client.get_hashes()

    not_in_client = set(torrents).difference(client_hashes)
    if len(not_in_client) != 0:
        log_stderr.info("Not in client:")
        for torrent_hash in not_in_client:
            log_stdout.print("%s", (torrents[torrent_hash].get_path(),))
        log_stderr.info("Found {red}%d{reset} unregistered torrents", (len(not_in_client),))
    else:
        log_stderr.info("No unregistered files found")


def print_missing_torrents(
    client: BaseClient,
    torrents_dir_path: str,
    name_filter: str,
    log_stdout: Log,
    log_stderr: Log,
) -> None:

    torrents = tcollection.by_hash(tcollection.load_from_dir(
        path=torrents_dir_path,
        name_filter=name_filter,
        precalculate_hashes=True,
        log=log_stderr,
    ))

    log_stderr.info("Fetching all hashes from client ...")
    client_hashes = client.get_hashes()

    missing_torrents = set(client_hashes).difference(torrents)
    if len(missing_torrents) != 0:
        log_stderr.info("Missing torrents for:")
        for torrent_hash in missing_torrents:
            log_stdout.print("%s -- %s", (torrent_hash, client.get_file_name(torrent_hash)))
        log_stderr.info("Found {red}%d{reset} torrents without torrent-files", (len(missing_torrents),))
    else:
        log_stderr.info("No torrents without torrent-files found")


def print_duplicate_torrents(
    torrents_dir_path: str,
    name_filter: str,
    log_stdout: Log,
    log_stderr: Log,
) -> None:

    torrents = {
        torrent_hash: variants
        for (torrent_hash, variants) in tcollection.by_hash_with_dups(tcollection.load_from_dir(
            path=torrents_dir_path,
            name_filter=name_filter,
            precalculate_hashes=True,
            log=log_stderr,
        )).items()
        if len(variants) > 1
    }
    if len(torrents) != 0:
        for (torrent_hash, variants) in torrents.items():
            log_stdout.print(torrent_hash)
            for torrent in variants:
                log_stdout.print("\t%s", (torrent.get_path(),))
    else:
        log_stderr.info("No duplicate torrents found")


# ===== Main =====
@wrap_main
def main() -> None:
    (parent_parser, argv, config) = init()
    args_parser = argparse.ArgumentParser(
        prog="emfind",
        description="Querying the client",
        parents=[parent_parser],
    )
    commands = args_parser.add_subparsers(dest="cmd")

    commands.add_parser(
        name="rebuild-cache",
        help="Rebuild files cache",
    )

    commands.add_parser(
        name="orphans",
        help="Reduce orphaned files when base directory also orphaned",
    ).add_argument("--no-reduce-dirs", action="store_true")

    commands.add_parser(
        name="not-in-client",
        help="Find torrent files, which are not registered in the client",
    )

    commands.add_parser(
        name="missing-torrents",
        help="Find torrents registered in the client for which there is no torrent files",
    )

    commands.add_parser(
        name="duplicate-torrents",
        help="Find torrent-files with duplicate hashes",
    )

    options = args_parser.parse_args(argv[1:])

    with get_configured_log(config, False, sys.stdout) as log_stdout:
        with get_configured_log(config, False, sys.stderr) as log_stderr:

            def get_client() -> BaseClient:
                return get_configured_client(  # type: ignore
                    config=config,
                    required=True,
                    with_customs=False,
                    log=log_stderr,
                )

            def get_cache(force_rebuild: bool) -> datacache.TorrentsCache:
                return datacache.get_cache(
                    cache_path=config.emfind.cache_file,
                    force_rebuild=force_rebuild,
                    client=get_client(),
                    torrents_dir_path=config.core.torrents_dir,
                    name_filter=config.emfind.name_filter,
                    log=log_stderr,
                )

            if options.cmd == "rebuild-cache":
                get_cache(True)

            elif options.cmd == "orphans":
                print_orphaned_files(
                    cache=get_cache(False),
                    data_roots=[config.core.data_root_dir] + config.core.another_data_root_dirs,
                    ignore_orphans=config.emfind.ignore_orphans,
                    reduce_dirs=(not options.no_reduce_dirs),
                    log_stdout=log_stdout,
                    log_stderr=log_stderr,
                )

            elif options.cmd in ["not-in-client", "missing-torrents"]:
                {
                    "not-in-client":    print_not_in_client,
                    "missing-torrents": print_missing_torrents,
                }[options.cmd](
                    client=get_client(),
                    torrents_dir_path=config.core.torrents_dir,
                    name_filter=config.emfind.name_filter,
                    log_stdout=log_stdout,
                    log_stderr=log_stderr,
                )

            elif options.cmd == "duplicate-torrents":
                print_duplicate_torrents(
                    torrents_dir_path=config.core.torrents_dir,
                    name_filter=config.emfind.name_filter,
                    log_stdout=log_stdout,
                    log_stderr=log_stderr,
                )


if __name__ == "__main__":
    main()  # Do the thing!



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import pkgutil
import textwrap

from typing import List
from typing import Dict
from typing import NamedTuple
from typing import Optional
from typing import Union
from typing import Type
from typing import Any

import mako.template

from ...optconf import Option

from ...plugins.trackers import BaseTracker  # FIXME

from ...tfile import TorrentsDiff
from ...tfile import Torrent

from .. import BasePlugin
from .. import get_classes


# =====
def templated(name: str, built_in: bool=True, **kwargs: Any) -> str:
    if built_in:
        data = pkgutil.get_data(__name__, os.path.join("templates", name))
        assert data, (data, name, __name__)
        text = data.decode()
    else:
        with open(name) as template_file:
            text = template_file.read()
    template = textwrap.dedent(text).strip()
    return mako.template.Template(template).render(**kwargs).strip()


# =====
class _InnerUpdateResult(NamedTuple):
    torrent: Optional[Torrent]
    tracker: Optional[BaseTracker]
    diff: TorrentsDiff
    err_name: str
    err_msg: str
    tb_lines: List[str]


class UpdateResult(_InnerUpdateResult):
    @staticmethod
    def new(
        torrent: Optional[Torrent]=None,
        tracker: Optional[BaseTracker]=None,
        diff: Optional[TorrentsDiff]=None,
        err_name: str="",
        err_msg: str="",
        tb_lines: Optional[List[str]]=None,
    ) -> "UpdateResult":

        return UpdateResult(
            torrent=torrent,
            tracker=tracker,
            diff=(diff or TorrentsDiff.new()),
            err_name=err_name,
            err_msg=err_msg,
            tb_lines=(tb_lines or []),
        )


ResultsType = Dict[str, Dict[str, UpdateResult]]


class BaseConfetti(BasePlugin):
    def __init__(self, timeout: float, retries: int, retries_sleep: float, **_: Any) -> None:  # pylint: disable=super-init-not-called
        self._timeout = timeout
        self._retries = retries
        self._retries_sleep = retries_sleep

    @classmethod
    def get_options(cls) -> Dict[str, Option]:
        return {
            "timeout":       Option(default=10.0, help="Network timeout"),
            "retries":       Option(default=5, help="Retries for failed attempts"),
            "retries_sleep": Option(default=1.0, help="Sleep interval between failed attempts"),
        }

    def send_results(self, source: str, results: ResultsType) -> None:
        raise NotImplementedError


class WithProxy(BaseConfetti):  # pylint: disable=abstract-method
    def __init__(self, proxy_url: str, **_: Any) -> None:  # pylint: disable=super-init-not-called
        self._proxy_url = proxy_url

    @classmethod
    def get_options(cls) -> Dict[str, Option]:
        return {
            "proxy_url": Option(default="", help="URL of HTTP/SOCKS4/SOCKS5 proxy"),
        }


def get_confetti_class(name: str) -> Type[BaseConfetti]:
    return get_classes("confetti")[name]  # type: ignore



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


from urllib.request import BaseHandler
from urllib.request import Request

from urllib.response import addinfo  # type: ignore
from urllib.response import addinfourl

import gzip

from typing import Union


# =====
class GzipHandler(BaseHandler):
    def http_request(self, request: Request) -> Request:
        request.add_header("Accept-Encoding", "gzip")
        return request

    def http_response(self, request: Request, response: addinfo) -> Union[addinfo, addinfourl]:  # pylint: disable=unused-argument
        if response.headers.get("Content-Encoding") == "gzip":
            gzip_file = gzip.GzipFile(fileobj=response, mode="r")
            new_response = addinfourl(gzip_file, response.headers, response.url, response.code)  # type: ignore
            new_response.msg = response.msg  # type: ignore
            return new_response
        return response

    https_request = http_request
    https_response = http_response



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import os
import argparse

from typing import List
from typing import Dict

from ..plugins.clients import BaseClient
from ..plugins.clients import WithCustoms

from ..helpers import tcollection

from ..tfile import Torrent

from .. import fmt

from . import init
from . import wrap_main
from . import get_configured_log
from . import get_configured_client


# =====
def make_path(path: str, tail_mode: int) -> None:
    try:
        os.makedirs(path)
        if tail_mode > 0:
            os.chmod(path, tail_mode)
    except FileExistsError:
        pass


def link_data(
    torrent: Torrent,
    data_dir_path: str,
    link_to_path: str,
    mkdir_mode: int,
) -> None:

    mkdir_path = link_to_path = os.path.abspath(link_to_path)
    if torrent.is_single_file():
        link_to_path = os.path.join(link_to_path, torrent.get_name())
    else:
        mkdir_path = os.path.dirname(link_to_path)

    if os.path.exists(link_to_path):
        raise RuntimeError("{}: link target already exists".format(link_to_path))

    make_path(mkdir_path, mkdir_mode)
    os.symlink(os.path.join(data_dir_path, torrent.get_name(surrogate_escape=True)), link_to_path)


def load_torrents(
    torrents: List[Torrent],
    client: BaseClient,
    data_root_path: str,
    link_to_path: str,
    mkdir_mode: int,
    customs: Dict[str, str],
) -> None:

    for torrent in torrents:
        if client.has_torrent(torrent):
            raise RuntimeError("{}: already loaded".format(torrent.get_path()))

    if not data_root_path:
        data_root_path = client.get_data_prefix_default()

    for torrent in torrents:
        dir_name = os.path.basename(torrent.get_path()) + ".data"
        data_dir_path = os.path.join(data_root_path, dir_name[0], dir_name)
        make_path(data_dir_path, mkdir_mode)

        if link_to_path:
            link_data(torrent, data_dir_path, link_to_path, mkdir_mode)

        client.load_torrent(torrent, data_dir_path)
        if WithCustoms in client.get_bases() and len(customs) != 0:
            client.set_customs(torrent, {  # type: ignore
                key: fmt.format_now(value)
                for (key, value) in customs.items()
            })


def parse_customs(items: List[str]) -> Dict[str, str]:
    customs = {}
    for item in filter(None, map(str.strip, items)):
        (key, value) = map(str.strip, (item.split("=", 1) + [""])[:2])
        customs[key] = value
    return customs


# ===== Main =====
@wrap_main
def main() -> None:
    (parent_parser, argv, config) = init()
    args_parser = argparse.ArgumentParser(
        prog="emload",
        description="Load torrent to client",
        parents=[parent_parser],
    )
    args_parser.add_argument("-l", "--link-to", default="", metavar="<path>")
    args_parser.add_argument("--set-customs", default=[], nargs="+", metavar="<key=value>")
    args_parser.add_argument("-v", "--verbose", action="store_true")
    args_parser.add_argument("torrents", nargs="+", metavar="<path>")
    options = args_parser.parse_args(argv[1:])

    if len(options.torrents) > 1 and options.link_to:
        raise RuntimeError("Option -l/--link-to be used with only one torrent")

    customs = dict(config.emload.set_customs)
    customs.update(parse_customs(options.set_customs))

    torrents = tcollection.find_torrents(config.core.torrents_dir, options.torrents)

    with get_configured_log(config, (not options.verbose), sys.stderr) as log_stderr:
        client: BaseClient = get_configured_client(  # type: ignore
            config=config,
            required=True,
            with_customs=bool(customs),
            log=log_stderr,
        )

        load_torrents(
            torrents=torrents,
            client=client,
            data_root_path=config.core.data_root_dir,
            link_to_path=options.link_to,
            mkdir_mode=config.emload.mkdir_mode,
            customs=customs,
        )


if __name__ == "__main__":
    main()  # Do the thing!



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import pickle

from typing import Dict
from typing import NamedTuple

from ..plugins.clients import BaseClient

from ..tfile import TorrentEntryAttrs

from ..cli import Log

from . import tcollection


# =====
class CacheEntryAttrs(NamedTuple):
    files: Dict[str, TorrentEntryAttrs]
    prefix: str


class TorrentsCache(NamedTuple):
    version: int
    torrents: Dict[str, CacheEntryAttrs]


def get_cache(
    cache_path: str,
    force_rebuild: bool,
    client: BaseClient,
    torrents_dir_path: str,
    name_filter: str,
    log: Log,
) -> TorrentsCache:

    cache = _read(cache_path, force_rebuild, log)
    if _update(cache, client, torrents_dir_path, name_filter, log):
        _write(cache, cache_path, log)
    return cache


# =====
def _read(path: str, force_rebuild: bool, log: Log) -> TorrentsCache:
    fallback = TorrentsCache(
        version=0,
        torrents={},
    )
    if force_rebuild or not os.path.exists(path):
        return fallback

    log.info("Reading the cache from {cyan}%s{reset} ...", (path,))
    with open(path, "rb") as cache_file:
        try:
            cache_low = pickle.load(cache_file)
            if cache_low["version"] != fallback.version:
                return fallback
            else:
                return TorrentsCache(
                    version=cache_low["version"],
                    torrents=pickle.loads(cache_low["torrents_pk"]),
                )
        except (KeyError, ValueError, pickle.UnpicklingError):
            log.error("Can't unpickle cache file - ingored: {red}%s{reset}", (path,))
            return fallback


def _write(cache: TorrentsCache, path: str, log: Log) -> None:
    os.makedirs(os.path.dirname(path), exist_ok=True)
    log.info("Writing the cache to {cyan}%s{reset} ...", (path,))
    with open(path, "wb") as cache_file:
        pickle.dump({
            "version": cache.version,
            "torrents_pk": pickle.dumps(cache.torrents),
        }, cache_file)


def _update(cache: TorrentsCache, client: BaseClient, path: str, name_filter: str, log: Log) -> bool:
    log.info("Fetching all hashes from client ...")
    hashes = client.get_hashes()

    log.info("Validating the cache ...")

    # --- Old ---
    to_remove = sorted(set(cache.torrents).difference(hashes))
    if len(to_remove) != 0:
        for torrent_hash in to_remove:
            cache.torrents.pop(torrent_hash)
        log.info("Removed {magenta}%d{reset} obsolete hashes from cache", (len(to_remove),))

    # --- New ---
    to_add = sorted(set(hashes).difference(cache.torrents))
    added = 0
    if len(to_add) != 0:
        torrents = tcollection.by_hash(tcollection.load_from_dir(path, name_filter, True, log))

        if not log.isatty():
            log.info("Adding files for the new {yellow}%d{reset} hashes ...", (len(to_add),))

        for torrent_hash in log.progress(
            to_add,
            ("Adding files ...", ()),
            ("Added {magenta}%d{reset} new hashes from client", (lambda: added,))
        ):
            torrent = torrents.get(torrent_hash)
            if torrent is not None:
                cache.torrents[torrent_hash] = CacheEntryAttrs(
                    files=torrent.get_files(),
                    prefix=client.get_data_prefix(torrent),
                )
                added += 1
            else:
                log.error("Not cached - missing torrent for: {red}%s{reset} -- %s",
                          (torrent_hash, client.get_file_name(torrent_hash)))

        if not log.isatty() and added != 0:
            log.info("Added {magenta}%d{reset} new hashes from client", (added,))

    return bool(len(to_remove) or added)



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import socket
import urllib.parse

from urllib.request import HTTPHandler
from urllib.request import HTTPSHandler
from urllib.request import Request

from http.client import HTTPConnection
from http.client import HTTPSConnection
from http.client import HTTPResponse

from typing import Tuple
from typing import Optional
from typing import Any

from ..thirdparty import socks


# =====
SCHEME_TO_TYPE = {
    "socks4": socks.PROXY_TYPE_SOCKS4,
    "socks5": socks.PROXY_TYPE_SOCKS5,
}

SOCKS_PORT = 1080


# =====
class _SocksConnection(HTTPConnection):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs.pop("proxy_url", None)  # XXX: Fix for "TypeError: __init__() got an unexpected keyword argument 'proxy_url'"
        super().__init__(*args, **kwargs)
        self._proxy_args: Optional[Tuple[
            Optional[int],
            Optional[str],
            Optional[int],
            bool,
            Optional[str],
            Optional[str],
        ]] = None

    # XXX: because proxy args/kwargs break super
    def make_proxy_args(
        self,
        proxy_url: str="",
        proxy_type: Optional[int]=None,
        proxy_host: Optional[str]=None,
        proxy_port: Optional[int]=None,
        proxy_user: Optional[str]=None,
        proxy_passwd: Optional[str]=None,
        rdns: bool=True,
    ) -> None:

        if proxy_url:
            parsed = urllib.parse.urlparse(proxy_url)
            scheme = parsed.scheme
            proxy_user = parsed.username
            proxy_passwd = parsed.password
            proxy_host = parsed.hostname
            proxy_port = (parsed.port or SOCKS_PORT)
            proxy_type = SCHEME_TO_TYPE.get((scheme or "").lower())
            if proxy_type is None:
                raise RuntimeError("Invalid SOCKS protocol: {}".format(scheme))

        self._proxy_args = (proxy_type, proxy_host, proxy_port, rdns, proxy_user, proxy_passwd)

    def connect(self) -> None:
        assert self._proxy_args is not None, "Proxy args weren't initialized"
        self.sock = socks.socksocket()
        self.sock.setproxy(*self._proxy_args)
        if self.timeout is not socket._GLOBAL_DEFAULT_TIMEOUT:  # type: ignore  # pylint: disable=protected-access
            self.sock.settimeout(self.timeout)  # type: ignore
        self.sock.connect((self.host, self.port))  # type: ignore


class _SocksSecureConnection(HTTPSConnection, _SocksConnection):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs.pop("proxy_url", None)  # XXX: Fix for "TypeError: __init__() got an unexpected keyword argument 'proxy_url'"
        super().__init__(*args, **kwargs)


# =====
class SocksHandler(HTTPHandler, HTTPSHandler):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._args = args
        self._kwargs = kwargs
        super().__init__(debuglevel=kwargs.pop("debuglevel", 0))

    def http_open(self, req: Request) -> HTTPResponse:
        def build(
            host: str,
            port: Optional[int]=None,
            timeout: int=socket._GLOBAL_DEFAULT_TIMEOUT,  # pylint: disable=protected-access
        ) -> _SocksConnection:

            connection = _SocksConnection(host, port=port, timeout=timeout, **self._kwargs)
            connection.make_proxy_args(*self._args, **self._kwargs)
            return connection

        return self.do_open(build, req)  # type: ignore

    def https_open(self, req: Request) -> HTTPResponse:
        def build(
            host: str,
            port: Optional[int]=None,
            timeout: int=socket._GLOBAL_DEFAULT_TIMEOUT,  # pylint: disable=protected-access
        ) -> _SocksSecureConnection:

            connection = _SocksSecureConnection(host, port=port, timeout=timeout, **self._kwargs)
            connection.make_proxy_args(*self._args, **self._kwargs)
            return connection

        return self.do_open(build, req)  # type: ignore



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import re
import operator

from datetime import datetime

from typing import Tuple
from typing import List
from typing import Dict
from typing import Any

from dateutil.relativedelta import relativedelta

from ...optconf import Option

from ...tfile import Torrent

from . import WithLogin
from . import WithCheckTime


# =====
class Plugin(WithLogin, WithCheckTime):
    PLUGIN_NAME = "tr.anidub.com"

    _SITE_VERSION = 2

    _SITE_FINGERPRINT_URL = "https://tr.anidub.com"
    _SITE_FINGERPRINT_TEXT = ("<link rel=\"search\" type=\"application/opensearchdescription+xml\""
                              " href=\"https://tr.anidub.com/engine/opensearch.php\"")

    _COMMENT_REGEXP = re.compile(r"https?://tr\.anidub\.com/\?newsid=(?P<torrent_id>\d+)")

    _TIMEZONE_STATIC = "Etc/GMT+4"

    # ===

    def __init__(self, **kwargs: Any) -> None:  # pylint: disable=super-init-not-called
        self._init_bases(**kwargs)
        self._init_opener(with_cookies=True)

    @classmethod
    def get_options(cls) -> Dict[str, Option]:
        return cls._get_merged_options()

    def fetch_time(self, torrent: Torrent) -> int:
        self._assert_match(torrent)

        date = self._assert_logic_re_search(
            regexp=re.compile(r"<li><b>Дата:</b> ([^,\s]+, \d\d:\d\d)</li>"),
            text=self._decode(self._read_url(torrent.get_comment())),
            msg="Upload date not found",
        ).group(1)

        now = datetime.now(self._tzinfo)
        day_template = "{date.day:02d}-{date.month:02d}-{date.year}"
        if "Сегодня" in date:
            date = date.replace("Сегодня", day_template.format(date=now))
        if "Вчера" in date:
            yesterday = now - relativedelta(days=1)
            date = date.replace("Вчера", day_template.format(date=yesterday))
        date += " " + datetime.now(self._tzinfo).strftime("%z")

        upload_time = int(datetime.strptime(date, "%d-%m-%Y, %H:%M %z").strftime("%s"))
        return upload_time

    def fetch_new_data(self, torrent: Torrent) -> bytes:
        self._assert_match(torrent)
        page = self._decode(self._read_url(torrent.get_comment().replace("http://", "https://")))
        downloads = set(map(int, re.findall(r"<a href=\"/engine/download.php\?id=(\d+)\" class=\" \">", page)))
        candidates: Dict[str, List[Tuple[Torrent, str]]] = {}
        for download_id in downloads:
            data = self._assert_valid_data(self._read_url(
                url="https://tr.anidub.com/engine/download.php?id={}".format(download_id),
                headers={"Referer": torrent.get_comment()},
            ))
            candidate = Torrent(data=data)

            name = candidate.get_name()
            candidates.setdefault(name, [])
            candidates[name].append((candidate, str(download_id)))

        name = torrent.get_name()
        self._assert_logic(name in candidates, "Can't find torrent named '{}' in downloads".format(name))
        self._assert_logic(len(candidates[name]) == 1, "Too many variants to download: {}".format(
                                                       ", ".join(map(operator.itemgetter(1), candidates[name]))))
        return candidates[name][0][0].get_data()

    def login(self) -> None:
        self._login_using_post(
            url="https://tr.anidub.com/",
            post={
                "login_name":      self._encode(self._user),
                "login_password":  self._encode(self._passwd),
                "login":           b"submit",
            },
            ok_text="<li><a href=\"https://tr.anidub.com/user/{}/\">Мой профиль</a></li>".format(self._user)
        )



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import socket
import urllib.request
import urllib.parse
import urllib.error
import http.client
import http.cookiejar
import time

from typing import List
from typing import Dict
from typing import Type
from typing import Optional

from . import gziphandler
from . import sockshandler


# =====
def build_opener(
    proxy_url: str="",
    cookie_jar: Optional[http.cookiejar.CookieJar]=None,
) -> urllib.request.OpenerDirector:

    handlers: List[urllib.request.BaseHandler] = [gziphandler.GzipHandler()]

    if proxy_url:
        scheme = (urllib.parse.urlparse(proxy_url).scheme or "").lower()
        if scheme in ["http", "https"]:
            handlers.append(urllib.request.ProxyHandler({scheme: proxy_url}))
        elif scheme in ["socks4", "socks5"]:
            handlers.append(sockshandler.SocksHandler(proxy_url=proxy_url))
        else:
            raise RuntimeError("Invalid proxy protocol: {}".format(scheme))

    if cookie_jar is not None:
        handlers.append(urllib.request.HTTPCookieProcessor(cookie_jar))

    return urllib.request.build_opener(*handlers)


def read_url(
    opener: urllib.request.OpenerDirector,
    url: str,
    data: Optional[bytes]=None,
    headers: Optional[Dict[str, str]]=None,
    timeout: float=10.0,
    retries: int=10,
    retries_sleep: float=1.0,
    retry_codes: Optional[List[int]]=None,
    retry_timeout: bool=True,
) -> bytes:

    if retry_codes is None:
        retry_codes = [500, 502, 503]

    while True:
        try:
            request = urllib.request.Request(url, data, (headers or {}))
            return opener.open(request, timeout=timeout).read()
        except socket.timeout as err:
            if retries == 0 or not retry_timeout:
                raise
        except urllib.error.HTTPError as err:
            if retries == 0 or err.code not in retry_codes:
                raise
        except urllib.error.URLError as err:
            if "timed out" in str(err.reason):
                if retries == 0 or not retry_timeout:
                    raise
            else:
                raise
        except (http.client.IncompleteRead, http.client.BadStatusLine, ConnectionResetError) as err:
            if retries == 0:
                raise

        time.sleep(retries_sleep)
        retries -= 1



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import os
import re
import time

from enum import Enum

from typing import TextIO
from typing import Tuple
from typing import List
from typing import NamedTuple
from typing import Iterable
from typing import Generator
from typing import Optional
from typing import Any

from colorama import Fore
from colorama import Style

from . import fmt


# =====
_COLORS = {
    "red":     Style.BRIGHT + Fore.RED,
    "green":   Style.BRIGHT + Fore.GREEN,
    "yellow":  Style.BRIGHT + Fore.YELLOW,
    "cyan":    Style.BRIGHT + Fore.CYAN,
    "magenta": Style.BRIGHT + Fore.MAGENTA,
    "blue":    Style.BRIGHT + Fore.BLUE,
    "reset":   Fore.RESET,
}

_NO_COLORS = dict.fromkeys(list(_COLORS), "")


# =====
class CellAlign(Enum):
    LEFT = "ljust"
    RIGHT = "rjust"
    CENTER = "center"


class Cell(NamedTuple):
    data: str
    colors: str = ""  # {red}
    align: CellAlign = CellAlign.LEFT


class Log:
    def __init__(
        self,
        use_colors: bool=True,
        force_colors: bool=False,
        quiet: bool=False,
        output: TextIO=sys.stdout,
    ) -> None:

        self._quiet = quiet
        self._output = output
        self._fill = 0
        self._colored = (use_colors and (self.isatty() or force_colors))
        self._ansi_regexp = re.compile(r"(\x1b[^m]*m)")

    def isatty(self) -> bool:
        return self._output.isatty()

    def info(self, text: str, placeholders: Tuple=(), one_line: bool=False, no_nl: bool=False) -> None:
        self.print("# {green}I{reset}: " + text, placeholders, one_line, no_nl)

    def error(self, text: str, placeholders: Tuple=(), one_line: bool=False, no_nl: bool=False) -> None:
        self.print("# {red}E{reset}: " + text, placeholders, one_line, no_nl)

    def print(self, text: str="", placeholders: Tuple=(), one_line: bool=False, no_nl: bool=False) -> None:
        if not self._quiet:
            rendered = self._format_text(text, placeholders, self._colored)
            view_len = len(self._format_text(text, placeholders, False) if self._colored else rendered)

            cut = 0
            if one_line and self.isatty():
                max_len = self._get_term_width()
                if max_len is not None and view_len > max_len:
                    cut = view_len - max_len
                    rendered = self._cut_line(rendered, cut)

            if view_len - cut < self._fill:
                self.finish()

            self._output.write(rendered)
            if one_line and self.isatty():
                self._output.write("\r")
                self._fill = view_len - cut
            elif not no_nl:
                self._output.write("\n")
                self._fill = 0
            self._output.flush()

    def print_table(self, header: List[Cell], table: List[List[Cell]]) -> None:
        assert len(header) >= 1
        widths = [len(cell.data) for cell in header]
        for row in table:
            for (cell_index, cell) in enumerate(row):
                if len(cell.data) > widths[cell_index]:
                    widths[cell_index] = len(cell.data)

        text_rows: List[str] = []
        for (row_index, row) in enumerate([header] + table):
            text_cells: List[str] = []
            for (cell_index, cell) in enumerate(row):
                text_cells.append(
                    cell.colors
                    + getattr(cell.data, cell.align.value)(widths[cell_index])
                    + ("{reset}" if cell.colors else "")
                )
            text_rows.append(" " + " | ".join(text_cells))

            if row_index == 0:
                text_rows.append("=" + "+".join(
                    "=" * (width + (2 if 0 < width_index < len(widths) else 1))
                    for (width_index, width) in enumerate(widths)
                ))

        self.print("\n".join(text_rows))

    def progress(
        self,
        iterable: Iterable[Any],
        wip: Tuple[str, Tuple],
        finish: Tuple[str, Tuple],
        length: int=20,
        refresh: float=0.1,
    ) -> Generator[Any, None, None]:

        if self.isatty():
            items = list(iterable)

            current = 0
            prev = 0.0
            for (current, item) in enumerate(items, 1):
                now = time.time()
                if prev + refresh < now:
                    (pb, pb_placeholders) = fmt.format_progress_bar(current, len(items), length)
                    self.info("{} :: {}".format(pb, wip[0]), pb_placeholders + wip[1], one_line=True)
                    prev = now
                yield item

            (pb, pb_placeholders) = fmt.format_progress_bar(current, len(items), length)
            self.info("{} :: {}".format(pb, finish[0]), pb_placeholders + finish[1])

        else:
            yield from iterable

    def finish(self) -> None:
        if self._fill:
            self._output.write((" " * self._fill) + "\r")
            self._fill = 0
            self._output.flush()

    def _get_term_width(self) -> Optional[int]:
        try:
            return int(os.environ["COLUMNS"])
        except (KeyError, ValueError):
            try:
                return os.get_terminal_size(self._output.fileno()).columns
            except OSError:
                return None

    def _format_text(self, text: str, placeholders: Tuple, colored: bool) -> str:
        text = text.format(**(_COLORS if colored else _NO_COLORS))
        text = text % tuple(
            (placeholder() if callable(placeholder) else placeholder)
            for placeholder in placeholders
        )
        return text

    def _cut_line(self, text: str, cut: int) -> str:
        parts = self._ansi_regexp.split(text)
        for index in reversed(range(len(parts))):
            if cut <= 0:
                break
            if parts[index].startswith("\x1b"):
                parts[index] = ""
            elif cut <= len(parts[index]):
                parts[index] = parts[index][:-cut]
                break
            else:
                cut -= len(parts[index])
                parts[index] = ""
        if self._colored:
            parts.append(Fore.RESET)
        return "".join(parts)



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import re
import hashlib
import base64
import urllib.parse
import itertools

from typing import List
from typing import Dict
from typing import Set
from typing import NamedTuple
from typing import Optional
from typing import Union
from typing import Any

import chardet

from .thirdparty import bencoder  # type: ignore


# =====
class _InnerTorrentEntryAttrs(NamedTuple):
    is_dir: bool
    size: int


class TorrentEntryAttrs(_InnerTorrentEntryAttrs):
    @staticmethod
    def new_file(size: int) -> "TorrentEntryAttrs":
        return TorrentEntryAttrs(is_dir=False, size=size)

    @staticmethod
    def new_dir() -> "TorrentEntryAttrs":
        return TorrentEntryAttrs(is_dir=True, size=0)


class _InnerTorrentsDiff(NamedTuple):
    added: Set[str]
    removed: Set[str]
    modified: Set[str]
    type_modified: Set[str]


class TorrentsDiff(_InnerTorrentsDiff):
    @staticmethod
    def new(
        added: Optional[Set[str]]=None,
        removed: Optional[Set[str]]=None,
        modified: Optional[Set[str]]=None,
        type_modified: Optional[Set[str]]=None,
    ) -> "TorrentsDiff":

        return TorrentsDiff(
            added=(added or set()),
            removed=(removed or set()),
            modified=(modified or set()),
            type_modified=(type_modified or set()),
        )


class Torrent:
    def __init__(self, data: Optional[bytes]=None, path: Optional[str]=None) -> None:
        # https://wiki.theory.org/index.php/BitTorrentSpecification

        self._path = path
        self._data: Optional[bytes] = None
        self._bencode: Optional[Dict] = None
        self._hash: str = ""
        self._scrape_hash: str = ""

        if data is not None:
            self.load_from_data(data, path)
        elif path is not None:
            self.load_from_file(path)

    def load_from_file(self, path: str) -> "Torrent":
        with open(path, "rb") as torrent_file:
            return self.load_from_data(torrent_file.read(), path)

    def load_from_data(self, data: bytes, path: Optional[str]=None) -> "Torrent":
        self._bencode = decode_torrent_data(data)
        self._path = path
        self._data = data
        self._hash = ""
        self._scrape_hash = ""
        return self

    # ===

    def get_path(self) -> str:
        assert self._path, self
        return self._path

    def get_data(self) -> bytes:
        assert self._data, self
        return self._data

    # ===

    def get_name(self, surrogate_escape: bool=False) -> str:
        assert self._bencode, (self, self._bencode)
        return self._decode(self._bencode[b"info"][b"name"], surrogate_escape)

    def get_comment(self) -> str:
        assert self._bencode, (self, self._bencode)
        return self._decode(self._bencode.get(b"comment", "").strip())

    def get_creation_date(self) -> int:
        assert self._bencode, (self, self._bencode)
        return self._bencode.get(b"creation date", 0)

    def get_created_by(self) -> Optional[str]:
        assert self._bencode, (self, self._bencode)
        created_by = self._bencode.get(b"created by")
        return (self._decode(created_by) if created_by is not None else None)

    def get_announce(self) -> Optional[str]:
        assert self._bencode, (self, self._bencode)
        announce = self._bencode.get(b"announce")
        return (self._decode(announce) if announce is not None else None)

    def get_announce_list(self) -> List[List[str]]:
        assert self._bencode, (self, self._bencode)
        return [
            list(map(self._decode, announce_list))
            for announce_list in self._bencode.get(b"announce-list", [])
        ]

    def is_private(self) -> bool:
        assert self._bencode, (self, self._bencode)
        return bool(self._bencode[b"info"].get(b"private", 0))

    # ===

    def get_hash(self) -> str:
        if not self._hash:
            assert self._bencode, (self, self._bencode)
            self._hash = hashlib.sha1(bencoder.bencode(self._bencode[b"info"])).hexdigest().lower()
        return self._hash

    def get_scrape_hash(self) -> str:
        if not self._scrape_hash:
            torrent_hash = self.get_hash()
            for index in range(0, len(torrent_hash), 2):
                self._scrape_hash += "%{}".format(torrent_hash[index:index + 2])
        return self._scrape_hash

    def make_magnet(self, extras: Optional[List[str]]=None) -> str:
        extras = (extras or [])

        assert self._bencode, (self, self._bencode)
        # http://stackoverflow.com/questions/12479570/given-a-torrent-file-how-do-i-generate-a-magnet-link-in-python
        info_sha1 = hashlib.sha1(bencoder.bencode(self._bencode[b"info"]))
        info_digest = info_sha1.digest()
        b32_hash = base64.b32encode(info_digest)

        magnet = "magnet:?xt={}".format(urllib.parse.quote_plus("urn:btih:{}".format(b32_hash)))
        if "name" in extras:
            magnet += "&dn={}".format(urllib.parse.quote_plus(self.get_name()))
        if "trackers" in extras:
            announces = self.get_announce_list()
            announce = self.get_announce()
            if announce:
                announces.insert(0, [announce])
            for announce in set(itertools.chain.from_iterable(announces)):
                magnet += "&tr={}".format(urllib.parse.quote_plus(announce))
        if "size" in extras:
            magnet += "&xl={}".format(self.get_size())
        return magnet

    # ===

    def get_size(self) -> int:
        assert self._bencode, (self, self._bencode)
        if self.is_single_file():
            return self._bencode[b"info"][b"length"]
        else:
            size = 0
            for fstruct in self._bencode[b"info"][b"files"]:
                size += fstruct[b"length"]
            return size

    def is_single_file(self) -> bool:
        assert self._bencode, (self, self._bencode)
        return (b"files" not in self._bencode[b"info"])

    def get_files(self, prefix: str="") -> Dict[str, TorrentEntryAttrs]:
        assert self._bencode, (self, self._bencode)
        base = os.path.join(prefix, self.get_name())
        if self.is_single_file():
            return {base: TorrentEntryAttrs.new_file(self._bencode[b"info"][b"length"])}
        else:
            files = {base: TorrentEntryAttrs.new_dir()}
            for fstruct in self._bencode[b"info"][b"files"]:
                name = None
                for index in range(len(fstruct[b"path"])):
                    name = os.path.join(base, os.path.sep.join(map(self._decode, fstruct[b"path"][0:index + 1])))
                    files[name] = TorrentEntryAttrs.new_dir()
                assert name is not None
                files[name] = TorrentEntryAttrs.new_file(fstruct[b"length"])
            return files

    # ===

    def _decode(self, value: Any, surrogate_escape: bool=False) -> str:  # pylint: disable=inconsistent-return-statements
        assert self._bencode, (self, self._bencode)
        if isinstance(value, bytes):
            if surrogate_escape:
                # https://www.python.org/dev/peps/pep-0383
                return value.decode("ascii", "surrogateescape")

            for encoding in [self._bencode.get(b"encoding", b"utf-8").decode(), "cp1251"]:
                try:
                    return value.decode(encoding)
                except UnicodeDecodeError:
                    pass

            encoding = chardet.detect(value)["encoding"]
            assert encoding is not None, "Can't determine encoding for bytes string: '{}'".format(repr(value))
            return value.decode(encoding)
        else:
            return value


def is_valid_torrent_data(data: bytes) -> bool:
    try:
        decode_torrent_data(data)
        return True
    except ValueError:
        return False


def decode_torrent_data(data: bytes) -> Dict:
    try:
        result = bencoder.bdecode(data)
    except bencoder.BTFailure as err:
        raise ValueError from err
    if not isinstance(result, dict):
        raise ValueError("Toplevel structure must be a dict")
    return result


def is_torrent_hash(text: str) -> bool:
    return (re.match(r"[\da-fA-F]{40}", text) is not None)


def get_torrents_difference(
    old: Union[Torrent, Dict[str, TorrentEntryAttrs]],
    new: Union[Torrent, Dict[str, TorrentEntryAttrs]],
) -> TorrentsDiff:

    old_files = (old.get_files() if isinstance(old, Torrent) else old)
    new_files = (new.get_files() if isinstance(new, Torrent) else new)

    modified = set()
    type_modified = set()

    for path in set(old_files).intersection(set(new_files)):
        if old_files[path].is_dir != new_files[path].is_dir:
            type_modified.add(path)
        elif old_files[path].size != new_files[path].size:
            modified.add(path)

    return TorrentsDiff(
        added=set(new_files).difference(set(old_files)),
        removed=set(old_files).difference(set(new_files)),
        modified=modified,
        type_modified=type_modified,
    )
